package qora.transaction;

import java.math.BigDecimal;
import java.util.Arrays;

import com.google.common.primitives.Bytes;

import data.transaction.GenesisTransactionData;
import data.transaction.TransactionData;
import qora.account.PrivateKeyAccount;
import qora.crypto.Crypto;
import transform.TransformationException;
import transform.transaction.TransactionTransformer;

public class GenesisTransaction extends Transaction {

	public GenesisTransaction(TransactionData transactionData) {
		this.transactionData = transactionData;
	}

	// Processing

	/**
	 * Refuse to calculate genesis transaction signature!
	 * <p>
	 * This is not possible as there is no private key for the genesis account and so no way to sign data.
	 * <p>
	 * <b>Always throws IllegalStateException.</b>
	 * 
	 * @throws IllegalStateException
	 */
	@Override
	public byte[] calcSignature(PrivateKeyAccount signer) {
		throw new IllegalStateException("There is no private key for genesis transactions");
	}

	/**
	 * Generate genesis transaction signature.
	 * <p>
	 * This is handled differently as there is no private key for the genesis account and so no way to sign data.
	 * <p>
	 * Instead we return the SHA-256 digest of the transaction, duplicated so that the returned byte[] is the same length as normal transaction signatures.
	 * 
	 * @return byte[]
	 */
	private byte[] calcSignature() {
		try {
			byte[] digest = Crypto.digest(TransactionTransformer.toBytes(this.transactionData));
			return Bytes.concat(digest, digest);
		} catch (TransformationException e) {
			return null;
		}
	}

	/**
	 * Check validity of genesis transaction signature.
	 * <p>
	 * This is handled differently as there is no private key for the genesis account and so no way to sign/verify data.
	 * <p>
	 * Instead we compared our signature with one generated by {@link GenesisTransaction#calcSignature()}.
	 * 
	 * @return boolean
	 */
	@Override
	public boolean isSignatureValid() {
		return Arrays.equals(this.transactionData.getSignature(), this.calcSignature());
	}

	@Override
	public ValidationResult isValid() {
		GenesisTransactionData genesisTransaction = (GenesisTransactionData) this.transactionData;

		// Check amount is zero or positive
		if (genesisTransaction.getAmount().compareTo(BigDecimal.ZERO) == -1)
			return ValidationResult.NEGATIVE_AMOUNT;

		// Check recipient address is valid
		if (!Crypto.isValidAddress(genesisTransaction.getRecipient()))
			return ValidationResult.INVALID_ADDRESS;

		return ValidationResult.OK;
	}

	@Override
	public void process() {
		// TODO
		// this.save();

		// Set recipient's balance
		// TODO
		// this.recipient.setConfirmedBalance(Asset.QORA, this.amount);

		// Set recipient's reference
		// TODO
		// recipient.setLastReference(this.signature);
	}

	@Override
	public void orphan() {
		// TODO
		// this.delete();

		// Reset recipient's balance
		// TODO
		// this.recipient.deleteBalance(Asset.QORA);

		// Set recipient's reference
		// TODO
		// recipient.setLastReference(null);
	}

}
